PAGE 66,132
TITLE bday2003.asm: programma assembly 8086 per calcolo prossimo compleanno

COMMENT *
Questo programma consente di calcolare il numero di giorni
mancanti al prossimo compleanno, nonche' il numero di anni
che si compiranno. Il programma mostra:
- l'utilizzo dell'int 21h del DOS per immettere una stringa
alfanumerica (formato ASCII) da tastiera (funzione 0ah), per
stampare a video (09h), per richiedere alla macchina la data
odierna (2ah) e per restituire il controllo al DOS (4ch);
- la conversione diretta e inversa di numeri interi
da formato ASCII decimale a formato binario, con uso di
operatori logici e matematici e look-up tables;
- la generazione di messaggi d'errore;
- la definizione di macro e procedure locali, ed il passaggio
di parametri attraverso lo stack.
    Il programma genera un numero limitato di messaggi d'errore,
evitando di controllare diverse situazioni di errata immissione
dati (errori di formato), come ad es. "03/04/78" al posto di
"03/04/1978", o "32/12/1974" (32 non e' un giorno accettabile),
o "1978" (non specificati giorno e mese).
Peraltro, il programma gestisce opportunamente e senza generare
messaggi d'errore situazioni del tipo "=27/11-)1913%%", in cui
i tre dati numerici siano preceduti, inframmezzati o seguiti da
caratteri non numerici (purche' annoverati tra quelli ammissibili).
    Nota: il programma e' stato scritto non nella sua forma piu'
sintetica, per due motivi: (1) mostrare come si possano usare
istruzioni e/o modi di indirizzamento diversi per raggiungere
uno stesso obiettivo; (2) favorire il riuso di parti di codice
per risolvere problemi differenti. Ad esempio, il "parsing" della
stringa digitata ed il controllo dei caratteri alfanumerici ammessi
potrebbe servire per realizzare una semplice calcolatrice per
svolgere le operazioni aritmetiche fondamentali da linea di
comando.
    Rispetto alla versione birthday.asm, viene incluso il controllo
sugli anni bisestili per l'anno di nascita, quello in corso, e 
quello prossimo. Inoltre, sono aggiunti alcuni nuovi commenti al 
vecchio codice.

Autore: Carlo Colombo, per il corso di Calcolatori I a.a. 2003-2004
Basato su: birthday.asm, scritto per il corso a.a. 2001-2002
Ultima revisione: 10/11/2003, 12:30h
*

;-----------------------------------------------------------------
; macro

        ; a. macro che fanno uso di int 21h (trap del sistema operativo)

keybuff macro buffer            ; lettura di una stringa da tastiera in un buffer di memoria
        push dx                 ; (N.B. la stringa termina con ENTER; il primo byte del buffer
        push bx                 ; contiene il max numero di caratteri da leggere, ed il secondo
        push ax                 ; byte contiene il numero totale di caratteri letti,
        mov dx,offset buffer    ; escluso CR (=ENTER); i caratteri letti (l'ultimo e' CR)
        mov ah,0ah              ; cominciano dal terzo byte
        int 21h
        mov bl,buffer[1]        ; (questa istruzione scrive in strlen)
        add bl,2                ; (i primi due byte del buffer)
        and bx,00ffh
        mov buffer[bx],'$'      ; Sovrascritto a CR, ed usato per motivi di display
                                ; non viene conteggiato in strlen=buffer[1]
        pop ax
        pop bx
        pop dx
        endm

display macro buffer             ; Scrittura a video di una stringa in memoria (buffer)
        push dx                  ; (N.B. ogni stringa deve terminare con '$')
        push ax
        mov dx,offset buffer
        mov ah,09h
        int 21h
        pop ax
        pop dx
        endm

getdate macro date_word          ; Return values: cx=year (1980-2099),
        push dx                  ; dh=month, dl=day, al=day of week (00h=sunday)
        push cx
        push ax
        mov ah,2ah
        int 21h
        mov date_word[0],cx
        mov al,dh
        and ax,00FFh
        mov date_word[2],ax
        and dx,00FFh
        mov date_word[4],dx
        pop ax
        pop cx
        pop dx
        endm

        ; b. macro di conversione

convert_daymon_into_days macro buffer   ; Il risultato e' sovrascritto in buffer[2]
local accum_days,stop                   ; Queste sono "dummy labels"
        push cx
        push dx
        push ax
        push si
        mov ax,0
        mov cx,buffer[2]
        dec cx                          ; qui si accumulano solo i giorni dei mesi interamente trascorsi
        cmp cx,0
        je stop
        mov dh,0
        mov si,0
accum_days:
        mov dl,months_days[si]
        add ax,dx
        inc si
        loop accum_days
stop:
        add ax,buffer[4]                ; qui si somma il contributo del mese corrente
        mov buffer[2],ax
        pop si
        pop ax
        pop dx
        pop cx
        endm

convert_ax_to_printable_ASCII macro stringa
        push ax
        push bx
        push ax                    
        mov ax,offset stringa
        push ax                      ; N.B. il passaggio dei parametri a bin2ascii 
        mov ax,maxN                  ; (=ax da conv., offset stringa, maxN)
        push ax                      ; avviene attraverso lo stack
        call near ptr bin2ascii
        pop bx                       ; in bx il numero di caratteri della stringa
        mov stringa[bx],'$'          ; (per potere stampare)
        add sp,4                     ; si riaggiusta SP in modo da evitare due pop 
        pop bx                       ; di dati ormai inutili (=offset stringa, ax da conv.)
        pop ax
        endm


        ; c. macro di controllo dati immessi
                                                         
disambiguate_60 macro buffer1,buffer2                      ; Questa macro usa il valore contenuto in buffer1[4]
    local no_inc_buffer2,inc_buffer2,check_if_29_february  ; per stabilire se un eventuale 60 in buffer1[2] sia
        cmp buffer1[2],60                                  ; dovuto alla digitazione di un 29/02 o di un 01/03:        
        jl no_inc_buffer2                                  ; 60 <== 31/01 + 29/02 oppure 
        jg inc_buffer2                                     ; 60 <== 31/01 + 28/02 + 01/03
check_if_29_february:                                      ; Il valore buffer2[2] viene incrementato solo nel   
        cmp buffer1[4],29                                  ; secondo caso.                                     
        je no_inc_buffer2
inc_buffer2:
        inc buffer2[2]       
no_inc_buffer2:
        endm


;-----------------------------------------------------------------
; Definizione costanti
CR     EQU 13                 ; carriage return (ritorno carrello - prima colonna)
LF     EQU 10                 ; line feed (avanza d'una riga)
maxN   equ 16                 ; max numero di caratteri leggibili da tastiera

;-----------------------------------------------------------------
;
DATA SEGMENT PUBLIC 'DATA'                  ; definizione segmento dati

messaggio_0     db  "Dammi la tua data di nascita (gg/mm/aaaa): $"
messaggio_err   db  "Errore di formato: i separatori ammessi sono $"
mess_a          db  "Il tuo prossimo compleanno ($"
bday_string     db  maxN dup('!')
mess_c          db  " anni) e' tra $"
daysleft_string db  maxN dup('%')
mess_e          db  " giorni: auguri!$"
a_capo          db  CR,LF,'$'

buffer_in       label byte                    ; buffer_in e' caratterizzato da 3 campi:
maxlen_in       db  maxN                      ; maxlen_in, strlen_in, stringa_in
strlen_in       db  ?
stringa_in      db  maxN dup('?')
dec_weight      dw  1
num_count       dw  0
num_accum       dw  0
num_ready       db  0

aux_vect db '/'                              ; vettore dei caratteri non numerici ammessi
         db '*'
         db '('
         db ')'
         db ' '
         db 'x'
         db '='
         db ':'
aux_tail db '-'
         db '$'

date_buffer     dw  3 dup(?)                 ; anno, mese, giorno
bday_buffer     dw  3 dup(?)                 ; come sopra
diff_buffer     dw  2 dup(?)                 ; come sopra

months_days db 31,28,31,30,31,30,31,31,30,31,30,31   ; "trenta di' conta novembre, con april, giugno e settembre;
                                                     ; di ventotto ce n'e' uno, tutti gli altri ne han trentuno".
                                                     ; Questa struttura dati non contempla gli anni bisestili:
                                                     ; sara' il codice a dover gestire questo caso particolare


DATA ENDS

;-----------------------------------------------------------------
;
STACK SEGMENT STACK 'STACK'    ; definizione del segmento di stack
      DB      64 DUP('STACK')  ; lo stack e' riempito con la stringa 'stack'
                               ; per identificarlo meglio in fase di debug
TOS   LABEL WORD               ; identifica il top of stack
STACK ENDS                     ;

;-----------------------------------------------------------------
;
CSEG SEGMENT PUBLIC 'CODE'

birthday PROC FAR
      ASSUME CS:CSEG,DS:DATA,SS:STACK,ES:DATA;

        MOV AX, seg DATA                 ; necessari 2 trasferimenti per assegnare a DS
        MOV DS,AX                        ; l'indirizzo del segmento dati in modo esplicito
        MOV ES,AX                        ; Nota: DS e ES puntano allo stesso segmento dati

        display a_capo
        display messaggio_0              ; Display messaggio iniziale e
        keybuff buffer_in                ; immissione di una stringa da tastiera
        display a_capo

        ; 1. scansione e analisi di ogni carattere ASCII di stringa_in,
        ;    con controllo del formato

            mov cl,strlen_in             ; Ciclo (esterno): inizializzazione
            mov ch,0
            xor bx,bx                    ;
leggi:      mov al,stringa_in[bx]        ; Il carattere e' prelevato dal buffer sorgente
            cmp al,'0'                   ; Controllo sul carattere ASCII: si vede se e'
            jl no_cifra                  ; una cifra (se non lo e', si fa un secondo
                                         ; controllo per vedere se e' un carattere non
            cmp al,'9'                   ; numerico - separatore - ammesso)
            jle cifra                    ; (E' una cifra)

                                         ; Ciclo (interno) che controlla se AL contiene
                                         ; un carattere ammesso oppure no
no_cifra:       lea si,aux_tail          ; Nota: lea = load effective address: equivale a
                lea dx,aux_vect          ;       mov si, offset aux_tail
controlla:      cmp al,ES:[si]           ; Nota: segment override (ES al posto del default DS,
                je ammesso               ;   con allungamento di 1 byte del codice di macchina)
                dec si                   ; Nota: si scandisce il vettore a partire
                                         ; dall'ultimo elemento, e si esce dal loop
                cmp si,dx                ; con un confronto tra effective addresses
                jge controlla
                display messaggio_err    ; stampa del messaggio d'errore e uscita dal programma
                display aux_vect         ;
                jmp exit
ammesso:
cifra:      add bx,1                     ; Controllo di fine ciclo esterno
            dec cx                       ; queste due istruzioni equivalgono a loop leggi
            jnz leggi



        ; 2. Scansione della stringa da destra a sinistra al fine di convertire
        ;    in forma binaria tutti i numeri inclusi nella stringa

        sub cx,cx
        mov cl,strlen_in
        mov si,cx
        dec si
new_char:                                 ; si verifica se il nuovo carattere letto dalla stringa
        mov al,stringa_in[si]             ; sia una cifra o un separatore
        cmp al,'0'
        jl no_digit
        cmp al,'9'
        jg no_digit

digit:                                    ; In AL c'e' una cifra decimale (in ASCII)
        cmp num_ready,1
        je no_new_number
        mov num_ready,1
        mov num_accum,0                   ; Accumulatore per il numero (codifica posizionale, base 10)
        mov dec_weight,1                  ; Peso (potenza del 10) da moltiplicare per la cifra
        inc num_count                     ; Conteggio dei numeri presenti nella stringa. Se la stringa
no_new_number:                            ; e' ben formata, num_count varra' alla fine 3 (=a,m,g)
        and ax,000Fh                      ; conversione ASCII -> binario: si rimuove il 30H della codifica ASCII
        mul dec_weight                    ; Il risultato di [AX]*dec_weight va in DX:AX (notare gli oper. impliciti)
        add num_accum,ax                  ; N.B.: Se la stringa e' corretta, DX varra' qui sempre 0 (infatti, il
        mov ax,10                         ; numero max da convertire e' il valore dell'anno di nascita, sempre
        mul dec_weight                    ; rappresentabile coi 16 bit di AX).
        mov dec_weight,ax                 ; ossia dec_weight <- dec_weight * 10
        jmp close_iteration

no_digit:                                 ; In AL c'e' un carattere separatore; esso potrebbe essere
        cmp num_ready,0                   ; ad inizio stringa (num_ready=1, num_accum contiene il giorno),
        je close_iteration                ; a fine stringa (num_ready=0, num_accum non contiene dati significativi),
        mov num_ready,0                   ; o nel mezzo (num_ready=1, num_accum contiene l'anno o il mese)
        mov ax,num_accum
        mov bx,num_count
        dec bx                            ; La word da salvare e' all'offset (num_count-1)*2 di bday_buffer
        add bx,bx                         ; equivale a raddoppiare il contenuto di bx
        mov bday_buffer[bx],ax            ; Salvataggio di num_accum

close_iteration:
        dec si
        dec cx
        cmp cx,0
        je check_save_day
        jmp new_char
check_save_day:                           ; Il giorno viene salvato solo se non l'avevamo gia' fatto prima,
        cmp num_ready,0                   ; cioe' solo se il primo carattere della stringa e' una cifra.
        je compute_birthday               ; Questo salto si fa se il primo carattere della stringa NON e'
        mov ax,num_accum                  ; una cifra
        mov bx,num_count
        dec bx
        add bx,bx
        mov bday_buffer[bx],ax


        ; 3. Calcolo del prossimo compleanno
        ;

compute_birthday:
        getdate date_buffer
        mov ax,date_buffer[0]                 ; differenza in anni
        sub ax,bday_buffer[0]
        mov diff_buffer[0],ax

        convert_daymon_into_days bday_buffer  ; Calcolo del numero di giorni trascorsi dall'inizio dell'anno per
        convert_daymon_into_days date_buffer  ; la data specificata in buffer: i risultati sovrascritti in buffer[2]

        mov ax,date_buffer[0]
        call near ptr check_if_leap_year      ; controllo se anno bisestile
        cmp ax,0
        je noadd
current_year_is_leap_year:
        disambiguate_60 date_buffer,date_buffer                 
        disambiguate_60 bday_buffer,bday_buffer                                                                     
noadd:
        mov ax,bday_buffer[2]                 ; Se la differenza in giorni dall'inizio dell'anno tra la data
        sub ax,date_buffer[2]                 ; del compleanno e quella attuale e' negativa, allora il compleanno
        mov diff_buffer[2],ax                 ; dell'anno in corso e' gia' passato
        cmp ax,0
        jge prepara_risultati
deja:   add diff_buffer[2],365                ; (qui la diff e' minore di zero, dunque si somma: l'op. e' modulo 365)
        mov ax,date_buffer[0]
        inc ax                                ; si vede se l'anno prossimo e' bisestile
        call near ptr check_if_leap_year
        cmp ax,0
        je noinc
next_year_is_leap_year:
        disambiguate_60 bday_buffer,diff_buffer  
noinc:
        inc diff_buffer[0]

prepara_risultati:                            ; Conversione ASCII e salvataggio in memoria
        mov ax,diff_buffer[0]
        convert_ax_to_printable_ASCII bday_string
        mov ax,diff_buffer[2]
        convert_ax_to_printable_ASCII daysleft_string

stampa_risultati:                             ; Stampa a video
        display a_capo
        display mess_a
        display bday_string
        display mess_c
        display daysleft_string
        display mess_e
        display a_capo

exit:
        mov ah,4ch                       ; altro modo per restituire il controllo al DOS
        int 21h

birthday ENDP



; routine di conversione di un numero a 16 bit (il cui valore e' posto nello stack)
; in una stringa di N caratteri, il cui indirizzo e' anch'esso inserito nello stack
; prima della chiamata. Il valore di ritorno N rimpiazza il valore immesso nello stack
; dal chiamante, che indica la dimensione massima della stringa, maxN
bin2ascii proc near
        push ax
        push bx
        push cx
        push dx
        push si
        push di
        push bp
        pushf                              ; esegue il push della PSW

        mov bp,sp
        add bp,16                          ; rimuove l'effetto dei push qui sopra
        mov ax,SS:[bp+6]                   ; numero da convertire (e' sepolto sotto 3 words)
                                           ; N.B. qui SS: e' superfluo, perche' bp lavora per
                                           ; default nello stack
        mov bx,10                          ; Le istruzioni che seguono realizzano la codifica
        mov cx,0                           ; inversa rispetto a bin2ascii: qui si converte un
        mov dx,0                           ; numero binario in una stringa ASCII stampabile a video
again:  div bx                             ; in DX il resto, in AX il quoziente di [DX:AX]/[BX] (op. impl.)
        or dx,0030h                        ; (Il metodo usato e' quello delle divisioni successive:
        push dx                            ; si divide ripetutamente per la base (10), finche' non
        inc cx                             ; si raggiunge un quoziente 0. Per passare da una cifra decimale
        cmp ax,0                           ; rappr. in binario alla rappr. ASCII basta aggiungere 30h.)
        je save_result
        xor dx,dx
        jmp again
save_result:
        mov di,0
        mov bx,[bp+4]                      ; offset della stringa destinazione
cycle:  pop dx
        mov [bx][di],dl
        inc di
        loop cycle
        mov [bp+2],di                      ; numero di caratteri nella stringa destinazione

        popf
        pop bp
        pop di
        pop si
        pop dx
        pop cx
        pop bx
        pop ax

        ret

bin2ascii endp


; si verifica se l'anno in ax sia bisestile o no
; in caso positivo, si pone ax=1, altrimenti ax=0.
check_if_leap_year proc near
        and ax,0003h ; 0000000000000011B
        cmp ax,0     ; l'anno bisestile e' multiplo di 4, 
        je leap_year ; ossia e' del tipo xxxxxxxxxxxxxx00B
        mov ax,0
        ret
leap_year: 
        mov ax,1
        ret
check_if_leap_year endp


CSEG ENDS

END birthday                     ; il programma comincia all'indirizzo di birthday
